# Copyright (C) 2006 Eitan Isaacson <eitan@ascender.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


import time
import gnomevfs
import flickrlib
import urllib
import tempfile
import os
import threading
from PIL import Image
from math import sqrt
from gconf import client_get_default


class TemporaryHandle(object):
	def __init__(self, uri, tree):
		print 'TemporaryHandle', uri, tree
		self.uri = uri
		self.tree = tree
		fd, name = tempfile.mkstemp()
		#print 'NAME:', name
		self.name = name
		self.fhandle = os.fdopen(fd,'w')
	def close(self):
		print 'closing!'
		self.fhandle.close()
		set_name = self.uri.path.split('/')[1]
		for set in self.tree.getSets():
			if set.getName() == urllib.unquote(set_name):
				break
			set = None
		if not set:
			raise gnomevfs.NotFoundError
		set.addPhoto(self.name,self.uri.path)
		os.remove(self.name)

	def __getattribute__(self, name):
			if name in ('uri', 'fhandle', 'name', 'close','tree'):
				return object.__getattribute__(self, name)
			return getattr(self.fhandle, name)
		

class Handle(object):
	def __init__(self, uri, tree):
		#print 'handleinit', uri.__str__()
		self.name = uri.short_name
		self.mode = 'r'
		self.set = None
		self.photo = None 
		self.dir = None
		self.filepos = None
		self.data = None
		self.tree = tree
		self.uri = uri
		components = uri.path.strip('/').split('/')
		if not len(components) or not components[0]:
			return
		set_name = components.pop(0)
		for set in self.tree.getSets():
			if set.getName() == urllib.unquote(set_name):
				self.set = set
		if not self.set:
			#print 'Set does not exist', set_name
			raise gnomevfs.NotFoundError	
		try:
			photo_name = components.pop(0)
		except IndexError:
			return
		for photo in self.set.getPhotos():
			print urllib.unquote(photo_name.replace('.jpg','')).__repr__(), '--', photo.getName().__repr__()
			if urllib.unquote(photo_name.replace('.jpg','')).decode('utf-8') == photo.getName().decode('utf-8'):
				self.photo = photo
		if not self.photo:
			print 'NOT FOUND', uri
			raise gnomevfs.NotFoundError, uri

	def open_dir(self):
		if self.set is None:
			self.dir = self.tree.getSets()
		else:
			self.dir = self.set.getPhotos()
		return
				
	def read_dir(self, file_info):
		try:
			##print ">>read_dir<<", self.dir
			fileobj = self.dir.pop(0)
		except IndexError:
			raise gnomevfs.EOFError
		self.__get_file_info_from_obj(fileobj, file_info, self.tree.isSelf)

	def close_dir(self):
		self.dir = None
		##print ">>close_dir<<", self.dir

	@staticmethod
	def __get_file_info_from_obj(fileobj, file_info, isSelf):	
		if isSelf:
			file_info.uid = os.geteuid()
			file_info.gid = os.getgid()
		else:
			file_info.uid = 65534
			file_info.gid = 65534
			perms = 0
		if fileobj.getType() == 'photoset':
			file_info.mime_type = "x-directory/normal"
			file_info.type = gnomevfs.FILE_TYPE_DIRECTORY
			file_info.permissions = 0755
			file_info.inode = int(fileobj.getID())
		elif fileobj.getType() == 'photo':
			file_info.size = 1024
			oformat = fileobj.getOriginalFormat()
			if oformat == 'jpg':
				file_info.mime_type = "image/jpeg"
			elif oformat == 'png':
				file_info.mime_type = "image/png"
			elif oformat == 'gif':
				file_info.mime_type = "image/gif"				
			file_info.type = gnomevfs.FILE_TYPE_REGULAR
			file_info.permissions = 0644
			file_info.mtime = fileobj.getDateTaken()
			file_info.ctime = file_info.mtime
			file_info.atime = file_info.mtime
			file_info.inode = int(fileobj.getID())
			file_info.symlink_name = fileobj.getThumbnailURL()
		else:
			file_info.mime_type = "x-directory/normal"
			file_info.type = gnomevfs.FILE_TYPE_DIRECTORY
			file_info.permissions = 0755
		file_info.name = str(fileobj.getName())

	def get_file_info(self, file_info, options):
		if self.photo:
			self.__get_file_info_from_obj(self.photo, file_info, self.tree.isSelf)
			return
		elif self.set:
			self.__get_file_info_from_obj(self.set, file_info, self.tree.isSelf)
			return
		else:
			self.__get_file_info_from_obj(self.tree, file_info, self.tree.isSelf)
			return

	def set_file_info(self, file_info):
		if self.photo:
			self.photo.setName(file_info.name.replace('.jpg',''))
		elif self.set:
			self.set.setName(file_info.name)
			

	def unlink(self):
		if self.photo:
			try:
				self.set.removePhoto(self.photo)
			except ValueError:
				raise gnomevfs.NotFoundError
			except IndexError:
				raise gnomevfs.NotSupportedError, 'Can\'t remove last photo of set'
		else:
			raise gnomevfs.NotSupportedError

	def open(self, mode):
#		#print ">>open<<"
		self.filepos = 0
		self.data = urllib.urlopen(self.photo.getOriginalURL())

	def read(self, buffer, num_bytes):
		##print ">>read<<", self.data, num_bytes
		if self.data is None:
			raise gnomevfs.IOError
		chunk = self.data.read(num_bytes)
		if not len(chunk):
			raise gnomevfs.EOFError
		buffer[:len(chunk)] = chunk
		self.filepos += len(chunk)
		return len(chunk)

	def close(self):
		#print ">>close<<"
		self.photo = None
		self.data = None
		self.filepos = None


class flickr_method:
	def __init__(self, method_name, args):
		#print ">>flickr.__init__<<", method_name, args
		_flickr_api_key = '48197b3ff4fb6f9d1caa650cb05764dd'
		_flickr_ssecret = '96d5319a6da0f358'
		agent = flickrlib.FlickrAgent(_flickr_api_key,_flickr_ssecret)
		_flickr_token = agent._token
		_flickr_user = agent._user
		self.agent = flickrWrap(_flickr_api_key,_flickr_ssecret,_flickr_token,_flickr_user)
		self.flickr_tree = None
		self.just_uploaded = None
		self._lock = MyLock('getTree')
	def vfs_open_directory(self, uri, open_mode, context):
		#print ">>flickr.open_directory<<", uri, open_mode, context
		self._getTree(uri)
		handle = Handle(uri,self.flickr_tree)
		handle.open_dir()
		return handle

	def vfs_read_directory(self, handle, file_info, context):
		#print ">>flickr.read_directory<<", handle, file_info, context
		handle.read_dir(file_info)

	def vfs_close_directory(self, handle, context):
		#print ">>flickr.close_directory<<", handle, context
		handle.close_dir()

	def vfs_get_file_info(self, uri, file_info, options, context):
		#print ">>flickr.get_file_info<<", uri, file_info, options, context
		self._getTree(uri)
		Handle(uri,self.flickr_tree).get_file_info(file_info, options)
	def vfs_check_same_fs(self, uri_a, uri_b, context):
		#print ">>flickr.check_same_fs<<", uri_a, uri_b, context
		if uri_a.host_name == uri_b.host_name and uri_a.scheme == uri_b.scheme:
			return True
		
	def vfs_get_file_info_from_handle(self,	handle, file_info, options, context):
		#print ">>flickr.get_file_info_from_handle<<", handle, file_info, options, context
		handle.get_file_info(file_info, options)
	def vfs_read(self, handle, buffer, num_bytes, context):
		return handle.read(buffer, num_bytes)
	def vfs_find_directory(self, uri, kind, create_if_needed, find_if_needed, perm, context):
		print '>>flickr.find_directory<<'
		return None
	def vfs_create(self, uri, mode, exclusive, perm, context):
		print '>>flickr.vfs_create<<', uri, mode, exclusive, perm, context
		self._getTree(uri)
		if not mode & gnomevfs.OPEN_WRITE:
			print 'vfs_create: not supported' 
			raise gnomevfs.NotSupportedError
		try:
			Handle(uri,self.flickr_tree)
			if not exclusive:
				print 'vfs_create: overwriting'
				return TemporaryHandle(uri,self.flickr_tree)
			else:
				print 'failed'
		except gnomevfs.NotFoundError:
			#print 'vfs_create: new'
			return TemporaryHandle(uri,self.flickr_tree)

	def vfs_open(self, uri, mode, context):
		#print '>>flickr.vfs_open<<', uri, mode, context
		self._getTree(uri)
		if mode & gnomevfs.OPEN_READ:
			handle = Handle(uri,self.flickr_tree)
			handle.open(mode)
			return handle
		else:
			raise gnomevfs.NotSupportedError

	def vfs_close(self, handle, context):
		#print '>>flickr.vfs_close<<', handle, context
		if handle is None:
			raise gnomevfs.NotSupportedError
		if 'w' in handle.mode:
			# Upload file.
			handle.close()
		else:
			handle.close()

	def vfs_set_file_info(self, uri, file_info, mask, context):
		#print '>>flickr.vfs_set_file_info<<', uri, file_info, mask, context
		self._getTree(uri)
		try:
			Handle(uri,self.flickr_tree).set_file_info(file_info)
		except gnomevfs.NotFoundError:
			if self.flickr_tree.just_uploaded:
				uri2 = gnomevfs.URI(str(uri).replace(uri.short_path_name,self.flickr_tree.just_uploaded))	
				Handle(uri2,self.flickr_tree).set_file_info(file_info)
			else:
				raise gnomevfs.NotFoundError, uri
	
	def vfs_read(self, handle, buffer, num_bytes, context):
		#print '>>flickr.vfs_read<<', handle, num_bytes
		return handle.read(buffer, num_bytes)

	def vfs_is_local(self, uri):
		#print '>>flickr.vfs_is_local<<', uri
		return False
	def vfs_find_directory(self, uri, kind, create_if_needed, find_if_needed, perm, context):
		pass
	def vfs_write(self, handle, rbuffer, bytes, context):
		#print 'write' #, handle, type(rbuffer), bytes, context
		if handle is not None:
			handle.write(rbuffer[:bytes])
			return bytes
		raise gnomevfs.NotSupportedError
	
	def vfs_unlink(self, uri, context):
		#print '>>flickr.vfs_unlink<<', uri, context
		self._getTree(uri)
		Handle(uri,self.flickr_tree).unlink()
	def vfs_move(self, uri_old, uri_new, force_replace, context=None):
		self._getTree(uri_old)
		handle_old = Handle(uri_old,self.flickr_tree)
		handle_new = Handle(uri_new.parent,self.flickr_tree)
		if not handle_old.photo or not handle_new.set:
			raise gnomevfs.NotSupportedError
		if handle_new.set is handle_old.set:
			return
		handle_old.set.movePhoto(handle_old.photo,handle_new.set)
	## Not implemented..
	def vfs_seek(self, handle, whence, offset, context):
		print 'Unsupported: vfs_seek'
		raise gnomevfs.NotSupportedError
	def vfs_tell(self, handle):
		print 'Unsupported: vfs_tell'
		raise gnomevfs.NotSupportedError
	def vfs_truncate_handle(self, handle, length, context):
		print 'Unsupported: vfs_truncate_handle'
		raise gnomevfs.NotSupportedError
	def vfs_make_directory(self, uri, perm, context):
		print 'Unsupported: vfs_make_directory'
		raise gnomevfs.NotSupportedError
	def vfs_remove_directory(self, uri, context):
		print 'Unsupported: vfs_remove_directory'
		raise gnomevfs.NotSupportedError
	def vfs_truncate(self, uri, length, context):
		print 'Unsupported: vfs_truncate'
		raise gnomevfs.NotSupportedError
	def vfs_create_symbolic_link(self, uri, target_reference, context):
		print 'Unsupported: vfs_create_symbolic_link'
		raise gnomevfs.NotSupportedError
	def vfs_file_control(self, handle, operation, data, context):
		print 'Unsupported: vfs_file_control'
		raise gnomevfs.NotSupportedError
	def _getTree(self,uri):
		self._lock.acquire()
		if not self.flickr_tree or self.flickr_tree.getHost() != uri.host_name:
			try:
				print "Create Tree %s != %s" % (self.flickr_tree.getName().lower(),uri.host_name)
			except:
				print "From NULL"
			self.flickr_tree = flickrRoot(self.agent,username=uri.host_name)
		self._lock.release()
		
class flickrRoot:
	def __init__(self,agent,username=None):
		self.gconf_client = client_get_default()
		self.host = username
		self.lastListing = 0
		self.just_uploaded = None
		self.agent = agent
		self._children_lock = MyLock('Sets')
		if not username or username == agent.user:
			self.user = self.agent.call().flickr.test.login()
			self.isSelf = True
		else:
			self.user = agent.call().flickr.people.findByUsername(username=username)
			self.isSelf = False
		self.getSets()
	def getSets(self):
		self._children_lock.acquire()
		if time.time() - self.lastListing > 30:
			try:
				photosets = self.agent.call().flickr.photosets.getList(user_id=self.user['id'])['photoset']
			except KeyError:
				photosets = []
			self.lastListing = time.time()
			self.children = []
			if self.isSelf:
				self.children.append(photosetNode(self.agent,{'id':'1','title':'Unsorted','type':'photoset'},self))
			for set in photosets:
				self.children.append(photosetNode(self.agent,set,self))
		self._children_lock.release()
		return self.children[:]
	def removeSet(self,set):
		if self.getID() == '1':
			raise StandardError
		self._children_lock.acquire()
		self.children[0]._children_lock.acquire()
		self.children[0].children.extend(set.children)
		self.children[0].lastListing = 0
		self.children[0]._children_lock.release()
		self.children.remove(set)
		self.agent.call().flickr.photosets.delete(photoset_id=set.getID())
		self._children_lock.release()

		
	def getType(self):
		return u'flickr_root'
	def getName(self):
		return self.user['username'][0]['text']
	def getHost(self):
		return self.host

class flickrNode:
	def getID(self):
		return self._struct['id']
	def getName(self):
		try:
			return self._struct['title'][0]['text'].encode('utf-8')
		except TypeError:
			return self._struct['title'].encode('utf-8')
		except:
			return self._struct['id'].encode('utf-8')
	def getType(self):
		return str(self._struct['type'])
	
class photosetNode(flickrNode):
	def __init__(self, agent, set, parent):
		self.agent = agent
		self._struct = set
		self.lastListing = 0
		self.parent = parent
		self._children_lock = MyLock('photos')
	def getPhotos(self):
		self._children_lock.acquire()
		if time.time() - self.lastListing > 30:
			if self.getID() == '1':
				photos = self.agent.call().flickr.photos.getNotInSet(extras='original_format,date_taken,date_upload',per_page='500')['photo']
			else:
				photos = self.agent.call().flickr.photosets.getPhotos(photoset_id=self.getID(),extras='original_format,date_taken,date_upload')['photo']
			self.lastListing = time.time()
			self.children = []
			for photo in photos:
				self.children.append(photoNode(self.agent,photo, self))
		self._children_lock.release()
		return self.children[:]
	def getPhotoTitles(self):
		self._children_lock.acquire()
		titles = [photo.getTitle() for photo in self.children]
		self._children_lock.release()
		return titles
	def removePhoto(self,photo):
		self._children_lock.acquire()
		if len(self.children) < 2:
			raise IndexError, 'Can\'t remove last photo'
		self.children.remove(photo)
		self.agent.call().flickr.photos.delete(photo_id=photo.getID())
		self._children_lock.release()
	def addPhoto(self,filename,title):
		self._children_lock.acquire()
		photo = {'title':title,'type':'photo'}
		if self.parent.gconf_client.get_bool('/apps/gnickr/uploading/enable_economic'):
			pixels = self.parent.gconf_client.get_int('/apps/gnickr/uploading/total_pixels')
			quality = self.parent.gconf_client.get_int('/apps/gnickr/uploading/quality')
			im = Image.open(filename)
			w,h = im.size
			if w*h >= pixels:
				ratio = sqrt(float(pixels)/(w*h))
				im = im.resize((int(w*ratio),int(h*ratio)),Image.ANTIALIAS)
				im.save(filename,'JPEG',quality=quality)
		photo['id'] = self.agent.call().flickr.photos.upload(filename=filename,title=title)['text']
		if self.getID() != '1':
			self.agent.call().flickr.photosets.addPhoto(photoset_id=self.getID(),photo_id=photo['id'])
		#self.children.append(photoNode(photo, self, self.agent))
		self.lastListing = 0
		self._children_lock.release()
	def movePhoto(self,photo,target_set):
		self._children_lock.acquire()
		target_set._children_lock.acquire()
		if self.getID() != '1':
			self.agent.call().flickr.photosets.removePhoto(photo_id=photo.getID(),photoset_id=self.getID())
		if target_set.getID() != '1':
			self.agent.call().flickr.photosets.addPhoto(photo_id=photo.getID(),photoset_id=target_set.getID())
		self.children.remove(photo)
		target_set.children.append(photo)
		target_set._children_lock.release()
		self._children_lock.release()
	def setName(self,name):
		name = name.decode('utf-8')
		self.parent._children_lock.acquire()
		if not self._struct['title'] [0]['text']== name:
			self.agent.call().flickr.photosets.editMeta(photoset_id=self.getID(),title=name)
			self._struct['title'][0]['text'] = name
		self.parent._children_lock.release()
		
class photoNode(flickrNode):
	def __init__(self, agent, photo, parent):
		self.agent = agent
		self._struct = photo
		self.parent = parent
	def setName(self,name):
		if name in self.parent.getPhotoTitles():
			raise gnomevfs.FileExistsError
		name = name.decode('utf-8')
		if not self._struct['title'] == name:
			desc = self.agent.call().flickr.photos.getInfo(photo_id=self.getID())['description'][0]['text']
			self.agent.call().flickr.photos.setMeta(photo_id=self.getID(),title=name,description=desc)
			self._struct['title'] = name
	def getDateTaken(self):
		try:
			return long(flickrlib.convertDate(self._struct['datetaken']))
		except:
			return None
	def getTitle(self):
		return self._struct['title'].encode('utf-8')
	def getName(self):
		if self.parent.getPhotoTitles().count(self.getTitle()) > 1:
			return self.getID()
		else:
			return self.getTitle()
	def getOriginalFormat(self):
		return str(self._struct['originalformat'])
	def getOriginalURL(self):
		return str(flickrlib.photoURL(self._struct,size='o'))
	def getThumbnailURL(self):
		return str(flickrlib.photoURL(self._struct,size='t'))

class flickrWrap:
	""" Just a small wrapper to create a new FlickrAgent instance.
	It keeps threads happy when you don't use a global instance of FlickrAgent"""
	def __init__(self,key,secret,token,user):
		self.key = key
		self.secret = secret
		self.token = token
		self.user = user
	def call(self):
		try:
			response = flickrlib.FlickrAgent(self.key,self.secret,token=self.token,user=self.user)
		except flickrlib.FlickrError, e:
			if e.code == '99':
				raise gnomevfs.AccessDeniedError('Must authorize gnickr with flickr')
			elif e.code == '1':
				raise gnomevfs.NotFoundError
			else:
				raise e
		return response
		

class MyLock:
	'''This is here so it will be easy to put debug'''
	def __init__(self,str):
		self.lock = threading.RLock()
		self.str = str
	def acquire(self):
		return self.lock.acquire()
	def release(self):
		return self.lock.release()

class FlickrVFSError(Exception):
	def __init__(self, code, message):
		self.code = code
		self.message = message
		if int(code) == 1:
			raise gnomevfs.NotFoundError
		elif 96 <= int(code) <= 99:
			raise gnomevfs.AccessDeniedError
		elif int(code) == 105:
			raise gnomevfs.ServiceNotAvailableError
	def __str__(self):
		return '%s: %s' % (self.code, self.message)

flickrlib.FlickrError = FlickrVFSError
